#include "crossroads.h"

void choose_backend () {
    int backends[MAX_BACKEND], nbackends = 0, i, j, k,
	target_set = 0, flat_weights = 1, weights[MAX_BACKEND], *sel_weights,
	tot_weights, lo_val, hi_val, done;
    unsigned long long values[MAX_BACKEND], nbest;
    unsigned nclients;
#   ifndef HAVE_SRANDDEV    
    struct timeval tv;
#   endif    

    /* Check that we're allowed to accept at all. */
    if (activeservice->maxconnections &&
	servicereport->nclients >= activeservice->maxconnections) {
	warning ("Service %s: max clients %d exceeded",
		 activeservice->name,
		 activeservice->maxconnections);
	current_backend = -1;
	return;
    }

    /* Populate the array of selectable backends. */
    for (i = 0; i < activeservice->nbackend; i++) {
	/* Backend i is a candidate if:
	 * (a) it is available, AND
	 * (b) there's no max of connections, or the actual nr of connections
	 *     is below max
	 */
	if (servicereport->backendstate[i].avail == st_available &&
	    (activeservice->backend[i].maxconnections == 0 ||
	     activeservice->backend[i].maxconnections >
	     servicereport->backendstate[i].nclients)) {
	    msg ("Service %s: "
		 "candidate back end: %d (max clients %d, active %d, "
		 "state %s)",
		 activeservice->name,
		 i, activeservice->backend[i].maxconnections,
		 servicereport->backendstate[i].nclients,
		 state_to_string(servicereport->backendstate[i].avail));
	    backends[nbackends++] = i;
	}
    }

    /* No backends? Nogo. */
    if (!nbackends) {
	current_backend = -1;
	warning ("No active backends to select!");
	return;
    }

    /* Only one backend? Then it's always the first one,
     * whatever you do. */
    if (nbackends == 1) {
	current_backend = backends[0];
	servicereport->last_backend = current_backend;
	msg ("Service %s: only 1 backend to select (%d)",
	     activeservice->name, current_backend);
	return;
    }

    /* More than 1 backend available.. make a wise choice. */
    switch (activeservice->dispatchtype) {

    case ds_roundrobin:
	/* Find the currently used backend, go one forward. */
	for (i = 0; i < nbackends; i++)
	    if (backends[i] == servicereport->last_backend) {
		i++;
		i %= nbackends;
		current_backend = backends[i];
		servicereport->last_backend = current_backend;
		msg ("Service %s: chosen backend (roundrobin): %d",
		     activeservice->name, current_backend);
		return;
	    }
	/* None found.. try the first one (run  to momma). */
	current_backend = backends[0];
	servicereport->last_backend = current_backend;
	msg ("Service %s: "
	     "chosen backend (roundrobin, without historical info): %d",
	     activeservice->name, current_backend);
	return;

    case ds_random:
	/* Re-randomize. */
#	ifdef HAVE_SRANDDEV
	sranddev();
	msg ("Service %s: randomier seeded with randdev", activeservice->name);
#	else	
	gettimeofday (&tv, 0);
	srand ((unsigned) tv.tv_usec);
	msg ("Service %s: randomizer seeded with %u",
	     activeservice->name, (unsigned) tv.tv_usec);
#	endif
	
	/* First of all let's see if all the weights are the same. */
	tot_weights = 0;
	for (i = 0; i < nbackends; i++) {
	    values[i] = activeservice->backend[backends[i]].weight;
	    if (activeservice->backend[backends[i]].weight != 1)
		flat_weights = 0;
	}

	if (! flat_weights) {
	    /* Non-flat weight random balancing. Determine the chance
	     * to hit a given back end. First we populate the weights array.
	     * Eg imagine we have 4 back ends with weights 1,2,3,4. Then
	     * ultimately we need to get a selection array of indices:
	     * 0000111223, meaning that the chance to hit index 0 is 4/10
	     * and so on */

	    /* First of all we reverse the values vector, to get 4,3,2,1 */
	    done = 0;
	    while (!done) {
		done = 1;
		lo_val = 9999999;
		hi_val = 0;
		/* Find highest/lowest weights in values vector. */
		for (i = 0; i < nbackends; i++) {
		    if (!values[i])
			continue;
		    done = 0;
		    if (lo_val >= values[i])
			lo_val = values[i];
		    if (hi_val <= values[i])
			hi_val = values[i];
		}
		
		if (done)
		    break;

		/* Stick the values into the duplicate 'weights' array,
		   but the other way 'round. */
		for (i = 0; i < nbackends; i++)
		    if (values[i] == lo_val)
			weights[i] = hi_val;
		    else if (values[i] == hi_val)			
			weights[i] = lo_val;

		for (i = 0; i < nbackends; i++)
		    if (values[i] == lo_val) {
			tot_weights += hi_val;
			values[i] = 0;
		    } else if (values[i] == hi_val) {
			tot_weights += lo_val;
			values[i] = 0;
		    }
	    }
	    
	    /* The 'weights' vector is now the reverse of 'values'.
	     * Create a selection array. */
	    sel_weights = xmalloc (tot_weights * sizeof(int));
	    k = 0;
	    for (i = 0; i < nbackends; i++)
		for (j = 0; j < weights[i]; j++) {
		    sel_weights[k] = i;
		    k++;
		}

	    /* Debugging: 
	     * for (k = 0; k < tot_weights; k++)
	     *	 msg ("random: sel_weight[%d] = %d", k, sel_weights[k]);
	     */
	    
	    /* Select the next random back end. Algorithm is:
	     * i = rand() % tot_weights;
	     * j = sel_weights[i];
	     * k = backends[j];
	     * current_backend = k;
	     * msg ("i=%d j=%d, k=%d", i, j, k);
	     */
	    i = rand() % tot_weights;
	    current_backend = backends[sel_weights[i]];
	    servicereport->last_backend = current_backend;
	    free (sel_weights);
	    msg ("Service %s: "
		 "chosen backend (weighted random): %d at index %d",
		 activeservice->name, current_backend, i);
	    return;
	}

	/* ELSE: Choose a random one of the availables. */
	current_backend = backends[rand() % nbackends];
	servicereport->last_backend = current_backend;
	msg ("Service %s: chosen backend (flat-weight random): %d",
	     activeservice->name, current_backend);
	return;

    case ds_bysize:
	/* Fill the values */
	for (i = 0; i < nbackends; i++) {
	    /* When dispatch over #-connections is in place, then we
	     * compute using the averaged value. However if the averaged
	     * value is 0 (not yet computed), we take the true value. */
	    if (activeservice->dispatchover &&
		servicereport->backendstate[backends[i]].avg_nbytes)
		values[i] =
		    servicereport->backendstate[backends[i]].avg_nbytes *
		    activeservice->backend[backends[i]].weight;
	    else
		values[i] = servicereport->backendstate[backends[i]].nbytes *
		    activeservice->backend[backends[i]].weight;
	    msg ("Service %s: "
		 "by size weighing backend %d has weight %d, value %u"
		 " (bytes=%u, avgbytes=%u)",
		 activeservice->name, 
		 backends[i], 
		 activeservice->backend[backends[i]].weight,
		 (unsigned) values[i],
		 (unsigned) servicereport->backendstate[backends[i]].nbytes,
		 (unsigned)
		     servicereport->backendstate[backends[i]].avg_nbytes);
	}
	/* Get the backend that has transported the least bytes */
	for (i = 0; i < nbackends; i++)
	    if (!target_set++ || values[i] < nbest) {
		nbest = values[i];
		current_backend = backends[i];
		servicereport->last_backend = current_backend;
		msg ("Service %s: "
		     "by size weighing: best so far is %d (value %g)",
		     activeservice->name, current_backend, nbest);
	    }
	msg ("Service %s: chosen backend (by size): %d",
	     activeservice->name, current_backend);
	return;

    case ds_byduration:
	/* Fill the values */
	for (i = 0; i < nbackends; i++) {
	    /* When dispatch over #-connections is in place, then we
	     * compute using the averaged value. However if the averaged
	     * value is 0 (not yet computed), we take the true value. */
	    if (activeservice->dispatchover &&
		servicereport->backendstate[backends[i]].avg_nsec)
		values[i] =
		    servicereport->backendstate[backends[i]].avg_nsec *
		    1000000 *
		    activeservice->backend[backends[i]].weight;
	    else
		values[i] = servicereport->backendstate[backends[i]].nsec *
		    1000000 *
		    activeservice->backend[backends[i]].weight;
	    msg ("Service %s: By duration weighing: backend %d has value %lu"
		 " (sec=%g, avgsec=%g, weight=%d)",
		 activeservice->name, 
		 backends[i], values[i],
		 servicereport->backendstate[backends[i]].nsec,
		 servicereport->backendstate[backends[i]].avg_nsec,
		 activeservice->backend[backends[i]].weight);
	}
	/* Get the backend that has transported the least connection secs */
	for (i = 0; i < nbackends; i++)
	    if (!target_set++ || values[i] < nbest) {
		nbest = values[i];
		current_backend = backends[i];
		servicereport->last_backend = current_backend;
		msg ("Service %s: "
		     "by duration weighing: best so far is %d (value %g)",
		     activeservice->name, current_backend, nbest);
	    }
	msg ("Service %s: chosen backend (by duration): %d",
	     activeservice->name, current_backend);
	return;

    case ds_byorder:
	/* Get the first available back end in line. */
	current_backend = backends[0];
	servicereport->last_backend = current_backend;
	msg ("Service %s: chosen backend (by order): %d",
	     activeservice->name, current_backend);
	return;

    case ds_byconnections:
	/* Get the weighing values. */
	for (i = 0; i < nbackends; i++) {
	    values[i] =
		servicereport->backendstate[backends[i]].nclients *
		activeservice->backend[backends[i]].weight;
	    msg ("Service %s: "
		 "by connections weighing: backend %d has value %u",
		 activeservice->name, backends[i], values[i]);
	}
	/* Find the minimum in the values array. */
	for (i = 0; i < nbackends; i++)
	    if (!target_set++ || values[i] < nclients) {
		nclients = values[i];
		current_backend = backends[i];
		servicereport->last_backend = current_backend;
		msg ("Service %s: "
		     "by connections weighing: best so far is %d (value %u)",
		     activeservice->name, current_backend, nclients);
	    }
	msg ("Service %s: chosen backend (by connections): %d",
	     activeservice->name, current_backend);
	return;

    default:
	/* Internal fry.. */
	error ("Service %s: internal error: unhandled dispatch type %d",
	       activeservice->name, activeservice->dispatchtype);
    }
}
