/* Includes. */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <math.h>
#include <pwd.h>
#include <setjmp.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/file.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>

/* flock() macros, incase they aren't present on your system. */
#ifndef LOCK_SH
#define LOCK_SH 1
#endif
#ifndef LOCK_EX
#define LOCK_EX 2
#endif
#ifndef LOCK_NB
#define LOCK_NB 4
#endif
#ifndef LOCK_UN
#define LOCK_UN 8
#endif

/* inet_addr() macros, incase they aren't present on your system. */
#ifndef INADDR_NONE
#define INADDR_NONE ((unsigned long) -1)
#endif

/* Types. */
typedef enum {				/* Config parsing related */
    cf_portspec,
    cf_verbosityspec,
    cf_backendspec,
    cf_serverspec,
    cf_dispatchspec,
    cf_revivespec,
    cf_shmkeyspec,
    cf_weightspec,
    cf_decayspec,
    cf_onsuccspec,
    cf_onfailspec,
    cf_connectiontimeoutspec,
    cf_maxconnectionsspec,
    cf_typespec,
    cf_dumpspec,
    cf_thruspec,
    cf_bindspec,
    cf_stickycookiespec,
    cf_setclientheaderspec,
    cf_addclientheaderspec,
    cf_appendclientheaderspec,
    cf_setserverheaderspec,
    cf_addserverheaderspec,
    cf_appendserverheaderspec,
} Conftype;

typedef union {				/* Integer of string value */
    int ival;				/* passed around by the parser */
    char *sval;
} IntOrString;

typedef struct {
    Conftype cf;
    IntOrString v;
} Confset;

typedef struct {
    int n;
    Confset *set;
} Confsets;

typedef enum {				/* Dispatching types */
    ds_roundrobin,
    ds_random,
    ds_bysize,
    ds_byduration,
    ds_byorder,
    ds_byconnections,
} Dispatchtype;

typedef enum {				/* Service types */
    type_any,
    type_http,
} Servicetype;

typedef struct {			/* Backend description */
    char *name;				/* .. back end identifier */
    int port;				/* .. TCP port */
    char *server;			/* .. server name */
    int verbosity;			/* .. debugging on/off */
    int weight;				/* .. weight in backend selection */
    int decay;				/* .. decay in % */
    char *onsuccess;			/* .. system() on success */
    char *onfailure;			/* .. or on failure */
    char *dumpfile;			/* .. traffic dump file */
    char *thruputfile;			/* .. traffic throughput file */
    int maxconnections;			/* .. max # of allowed connections */
    char *stickycookie;			/* .. cookie selector */
    char **addclientheader;		/* .. client hdrs to ADD */
    int naddclientheader;		/* .. table size */
    char **setclientheader;		/* .. client hdrs to SET */
    int nsetclientheader;		/* .. table size */
    char **appendclientheader;		/* .. client hdrs to APPEND */
    int nappendclientheader;		/* .. table size */
    char **addserverheader;		/* .. server hdrs to ADD */
    int naddserverheader;		/* .. table size */
    char **setserverheader;		/* .. server hdrs to SET */
    int nsetserverheader;		/* .. table size */
    char **appendserverheader;		/* .. server hdrs to APPEND */
    int nappendserverheader;		/* .. table size */
} Backend;

typedef struct {			/* Service description */
    char *name;				/* .. service name */
    char *bind;				/* .. address to bind to */
    unsigned port;			/* .. listening port */
    unsigned verbosity;			/* .. message generation */
    Dispatchtype dispatchtype;		/* .. backend selection method */
    unsigned dispatchover;		/* .. selection over # connections */
    unsigned rev_interval;		/* .. dead backend revival interval */
    unsigned backlog;			/* .. # pending TCP connections */
    unsigned shmkey;			/* .. key for SysV IPC */
    unsigned connectiontimeout;		/* .. # secs for timeout handling */
    unsigned maxconnections;		/* .. max # of allowed connections */
    Servicetype type;			/* .. type of the service */
    Backend *backend;			/* .. the back ends */
    int nbackend;			/* .. size of backend array */
} Service;

typedef enum {				/* Backend availability */    
    st_available,			/* .. all OK */
    st_unavailable,			/* .. temporarily unavailable */
    st_down,				/* .. permanently unavailable */\
    st_waking,				/* .. scheduled for wake up */
    st_intermediate,			/* .. bogus, for mark_activity() */
    st_unknown,				/* .. no idea (terminator) */
} Backendavail;

typedef struct {			/* Backend state */
    Backendavail avail;			/* .. availability */
    unsigned long totuses;		/* .. times hit */
    unsigned long failures;		/* .. times failed */
    unsigned long long nbytes;		/* .. transferred bytes */
    unsigned long sessions;		/* .. # of http sessions */
    unsigned long avg_nbytes;		/* .. averaged over # connections */
    double nsec;			/* .. connection durations */
    double avg_nsec;			/* .. averaged over # connections */
    unsigned nclients;			/* .. active clients */
} Backendstate;

typedef struct {			/* Service reporting (shmem) */
    int pid;				/* .. PID of service daemon */
    int rev_pid;			/* .. PID of revival handler */
    int nclients;			/* .. active clients */
    int last_backend;			/* .. last used back end */
    Backendstate			/* .. states of the back ends */
        backendstate[MAX_BACKEND];
} Servicereport;

typedef enum {				/* Stage of the program: */
    stage_main,				/* .. commandline */
    stage_waiting,			/* .. waiting for connections */
    stage_serving,			/* .. servicing a connection */
    stage_retrying,			/* .. waking up a backend */
} Programstage;				/* Update stagetostring.c when */
					/* modifying! */

typedef struct {			/* An HTTP Message */
    char **header;			/* Headers table, excluding first */
    int nheader;			/* Headers table size */
} HttpHeader;

typedef enum {				/* Time representation format */
    tm_localtime,			/* .. local */
    tm_gmtime,				/* .. UTC */
} TmType;

typedef enum {				/* TCP copying direction */
    dir_client_to_server,
    dir_server_to_client,
} CopyDirection;

typedef enum {				/* HTTP version */
    con_unknown,
    con_keepalive,
    con_close,
} HttpConnectionType;

/* Globals. */
#ifndef EXTERN
#define EXTERN extern
#endif

EXTERN Service *activeservice;		/* target service of a daemon */
EXTERN unsigned char *clbuf;		/* client socket input buffer */
EXTERN unsigned clbufpos, clbufmax;	/* .. position & bytes */
EXTERN char *client_ip;			/* connected client */
EXTERN char *config_file;		/* config to parse */
EXTERN int current_backend;		/* of a given service */
EXTERN int daemonized;			/* are we forked off yet */
EXTERN int flag_verbose;		/* flag: verbosity */
EXTERN int iflag_present;		/* flag: -i present */
EXTERN int interrupted;			/* got a signal? */
EXTERN char *laststring;		/* semantic lexer value */
EXTERN int listen_sock;			/* servicer listening socket */
EXTERN int log_activity;		/* log activy to syslog? */
EXTERN int log_facility;		/* openlog(3) facility, -l flag */
EXTERN int logstarted;			/* was syslog() called yet? */
EXTERN int nservice;			/* # service descriptions */
EXTERN int org_argc;			/* original argc */
EXTERN char **org_argv;			/* and original argv */
EXTERN Programstage program_stage;	/* stage of the program */
EXTERN int relevant_sigs[];		/* relevant signals */
EXTERN int semid;			/* semaphore ID */
EXTERN Service *service;		/* service descriptions */
EXTERN Servicereport *servicereport;	/* reporter in shared mem */
EXTERN unsigned char *srbuf;		/* server socket input buffer */
EXTERN unsigned srbufpos, srbufmax;	/* .. position & bytes */
EXTERN char *state_to_string_map[];	/* backend states as strings */
EXTERN FILE *yyin;			/* config file handle */
EXTERN int yylineno;			/* input line number */
EXTERN char *yyerrmsg;			/* parsing error msg */
EXTERN char *yytext;			/* lexical buffer */

/* Functions. */
extern void alloc_reporter (Service *active, int first);
extern char *ansistamp (TmType t);
extern int backend_available (void);
extern int backend_connect (void);
extern void choose_backend (void);
extern void copysockets (int clientsock, int serversock);
extern void configtest (int ac, char **av);
extern void decr_client_count (void);
extern void dealloc_reporter (Service *s);
extern void error (char const *fmt, ...);
extern int fork_tcp_servicer (int to_backend);
extern void http_copy (HttpHeader *h, int src_sock, int dst_sock,
		       CopyDirection dir);
extern void http_error (int clientsock);
extern char *http_expand_header (char const *h);
extern void http_header_addheader (HttpHeader *m, char const *hdr);
extern void http_header_appendheader (HttpHeader *m, char const *hdr);
extern HttpConnectionType http_header_connectiontype (HttpHeader *h);
extern void http_header_free (HttpHeader *msg);
extern int http_header_hascookie (HttpHeader *m, char const *cookie);
extern double http_header_httpver (HttpHeader *h);
extern void http_header_removeheader (HttpHeader *h, char const *hdr);
extern unsigned char const *http_header_val (HttpHeader *h, char const *hdr);
extern HttpHeader *http_header_new (void);
extern void http_header_read (HttpHeader *m, int srcsock, CopyDirection dir);
extern void http_header_setheader (HttpHeader *m, char const *hdr);
extern void http_header_write (HttpHeader *m, int sock, CopyDirection dir);
extern void http_serve (int clientsock);
extern int http_serversocket (HttpHeader *m, int *is_continuation);
extern int http_write (int sock, CopyDirection dir, unsigned char const *buf,
		       unsigned buflen);
extern void incr_client_count (void);
extern int is_hex_digit (char ch);
extern int is_space (char ch);
extern int init_sockaddr (struct sockaddr_in *name,
			   const char *hostname, int port);
extern void interrupt (int sig);
extern void lock_reporter (void);
extern void log_activity_any (char const *action);
extern void log_activity_start (void);
extern void log_activity_end (void);
extern void log_activity_continuation (void);
extern int make_socket (int port, char const *ipaddr);
extern void mark_activity (unsigned long long nbytes,
			   double nsec,
			   Backendavail newstate);
extern void msg (char const *fmt, ...);
extern void msgdumpbuf (unsigned char const *buf, int buflen);
extern unsigned net_copy (int clientsock, int serversock, unsigned maxbytes,
			  unsigned char *buf);
extern unsigned net_read (int sock, unsigned maxbytes, unsigned char *buf,
			  int is_client);
extern unsigned char *net_buffer (CopyDirection dir, unsigned *sz);
extern unsigned char *net_bufread (int sock, unsigned maxbytes,
				   unsigned *nreadp, int is_client);
extern int net_write (int sock, unsigned char const *buf, unsigned len,
		      int is_client);
extern void restart (int ac, char **av);
extern void runservice (void);
extern void sample_conf (int ac, char **av);
extern void serve (int ac, char **av);
extern void set_program_title (char const *fmt, ...);
extern void show_services (int ac, char **av);
extern void show_status (int ac, char **av);
extern char *stage_to_string (Programstage stage);
extern char *state_to_string (Backendavail avail);
extern Backendavail string_to_state (char const *str);
extern void stop_daemon (int ac, char **av);
extern char *str_printf (char const *fmt, ...);
extern char *str_vprintf (char const *fmt, va_list args);
extern void sysrun (char const *cmd);
extern void tcpserve (int sock);
extern void tell_service (int ac, char **av);
extern void thruputlog (unsigned char const *buf, int len, CopyDirection dir);
extern void trafficlog (unsigned char const *buf, int len, CopyDirection dir);
extern void unlock_reporter (void);
extern void usage (void);
extern void wakeup_handler (void);
extern void warning (char const *fmt, ...);
extern void writelog (int prio, char const *fmt, ...);
extern int yyparse (void);
extern void *xmalloc (unsigned sz);
extern void *xrealloc (void *what, unsigned sz);
extern char *xstrcat (char *what, char const *rest);
extern char *xstrcatch (char *what, char ch);
extern char *xstrdup (char const *what);

/* strlcat() if it's missing on this system */
#ifndef HAVE_STRLCAT
extern size_t strlcat (char *dst, char const *src, size_t size);
#endif

/* vsyslog() if it's missing on this system */
#ifndef HAVE_VSYSLOG
extern void vsyslog (int fac, char const *fmt, va_list args);
#endif
