#include "crossroads.h"

void http_serve (int clientsock) {
    int serversock = -1, is_continuation, i;
    HttpHeader *clientreq, *serverresp;
    int client_persisting, server_persisting;
    unsigned char *cp;
    unsigned size;

    /* Fork off a servicer if we're in daemon mode. */
    if (fork_tcp_servicer(-1))
	return;

    /*
     * Here's the logic of the HTTP services.
     * - Get the client's request;
     * - If we're not yet connected to a back end: determine the back
     *   end and connect to it
     * - Apply all server-directed headers
     * - Send client's request to the server
     * - Read back the server's response
     * - Apply all client-directed headers
     * - Send the server's response to the client
     * - Redo from start; this might be a Keep-Alive connection
     */

    while (1) {
	/* Assume both client and server want to persist. */
	client_persisting = server_persisting = 1;
	
	/* Get the initial client headers */
	clientreq = http_header_new();
	http_header_read (clientreq, clientsock, dir_client_to_server);
	
	/* Determine the back end if necessary. */
	if (serversock == -1) {
	    if ( (serversock = http_serversocket (clientreq,
						  &is_continuation)) < 1 )
		http_error (clientsock);
	    incr_client_count();
	    if (is_continuation)
		log_activity_continuation();
	    else
		log_activity_start ();
	}
	
	/* Put all server-directed headers in place. */
	for (i = 0;
	     i < activeservice->backend[current_backend].naddserverheader;
	     i++)
	    http_header_addheader (clientreq,
	        activeservice->backend[current_backend].addserverheader[i]);
	for (i = 0;
	     i < activeservice->backend[current_backend].nsetserverheader;
	     i++)
	    http_header_setheader (clientreq,
  	        activeservice->backend[current_backend].setserverheader[i]);
	for (i = 0;
	     i < activeservice->backend[current_backend].nappendserverheader;
	     i++)
	    http_header_appendheader (clientreq,
	        activeservice->backend[current_backend].appendserverheader[i]);

	/* See RFC2616. We're actually a proxy here. Clients that talk
	 * HTTP/1.0 MUST be treated as single shot connection clients, even
	 * when they specify Connection: Keep-Alive. */
	if (http_header_httpver (clientreq) < 1.1) {
	    msg ("Service %s: Client talks HTTP < 1.1, forcing closing "
		 "connections", activeservice->name);
	    http_header_setheader (clientreq, "Connection: close");
	    http_header_setheader (clientreq, "Proxy-Connection: close");
	    client_persisting = 0;
	} else if (http_header_connectiontype (clientreq) == con_close) {
	    msg ("Service %s: Client asks for connection closing",
		 activeservice->name);
	    client_persisting = 0;
	}

	/* Send client's headers to the back end. */
	http_header_write (clientreq, serversock, dir_client_to_server);
	
	/* Copy body from client to server */
	http_copy (clientreq, clientsock, serversock, dir_client_to_server);
	if ( (cp = net_buffer (dir_client_to_server, &size)) )
	    net_write (serversock, cp, size, 0);
	
	/* Get server response. */
	serverresp = http_header_new();
	http_header_read (serverresp, serversock, dir_server_to_client);

	/* According to RFC1616,  we should convert chunked output
	 * from the server to one big message.
	 * We won't do that here, it breaks streaming.
	 * But we will close connections in the case of chunked transfer. */
	if (client_persisting &&
	    http_header_val (serverresp, "transfer-encoding")) {
	    msg ("Service %s: Server is sending chunks, forcing singleshot",
		 activeservice->name);
	    server_persisting = 0;
	    http_header_setheader (serverresp, "Connection: close");
	    http_header_setheader (serverresp, "Proxy-Connection: close");
	}

	/* Put client-directed headers in place. */
	for (i = 0;
	     i < activeservice->backend[current_backend].naddclientheader;
	     i++)
	    http_header_addheader (serverresp,
	        activeservice->backend[current_backend].addclientheader[i]);
	for (i = 0;
	     i < activeservice->backend[current_backend].nsetclientheader;
	     i++)
	    http_header_setheader (serverresp,
	        activeservice->backend[current_backend].setclientheader[i]);
	for (i = 0;
	     i < activeservice->backend[current_backend].nappendclientheader;
	     i++)
	    http_header_appendheader (serverresp,
	        activeservice->backend[current_backend].appendclientheader[i]);

	/* According to RFC1616,  we should convert chunked output
	 * from the server to one big message.
	 * We won't do that here, it breaks streaming.
	 * But we will close connections in the case of chunked transfer. */
	if (client_persisting &&
	    http_header_val (serverresp, "transfer-encoding")) {
	    msg ("Service %s: Server is sending chunks, forcing singleshot",
		 activeservice->name);
	    server_persisting = 0;
	    client_persisting = 0;
	    http_header_setheader (serverresp, "Connection: close");
	    http_header_setheader (serverresp, "Proxy-Connection: close");
	}

	/* If the server wants to close the connection, then we must
	 * inform the client. */
	if (client_persisting) {
	    if (http_header_httpver (serverresp) < 1.1) {
		msg ("Service %s: Server talks HTTP < 1.1, forcing closing "
		     "connections", activeservice->name);
		http_header_setheader (serverresp, "Connection: close");
		http_header_setheader (serverresp, "Proxy-Connection: close");
		server_persisting = 0;
	    } else if (http_header_connectiontype (serverresp) == con_close) {
		msg ("Service %s: Server asks for connection closing",
		 activeservice->name);
		http_header_setheader (serverresp, "Connection: close");
		http_header_setheader (serverresp, "Proxy-Connection: close");
		server_persisting = 0;
	    }
	}
	
	/* Send server headers to the client. */
	http_header_write (serverresp, clientsock, dir_server_to_client);
	
	/* Copy body from server to cliet and flush network buffers if any */
	http_copy (serverresp, serversock, clientsock, dir_server_to_client);
	if ( (cp = net_buffer (dir_server_to_client, &size)) )
	    net_write (clientsock, cp, size, 0);

	/* Free up info from this request/response chat. */
	http_header_free (clientreq);
	http_header_free (serverresp);

	/* Should we loop around? Go to copy-thru mode if either the
	 * client or the server wants to stop. */
	if (!client_persisting || !server_persisting) {
	    msg ("Service %s: non-persistent connection",
		 activeservice->name);
	    break;
	}
	
	msg ("Service %s: Analyzing following requests over same TCP link",
	     activeservice->name);
    }

    /* Copy through all that arrives on back end or client.
     * This doesn't return... */
    msg ("Service %s: HTTP service done analyzing, entering copy-thru mode",
	 activeservice->name);

    /* This doesn't return */
    copysockets (clientsock, serversock);

}
