#include "crossroads.h"

void tcpserve (int server_sock) {
    fd_set set;
    int backend_sock, new, size, i, pid;
    struct sockaddr_in clientname;
    static int wakeup_started = 0;

    /* Set up our signal handlers. */
    if (!wakeup_started++) {
	/* Interruption signals */
	for (i = 0; relevant_sigs[i]; i++)
	    signal (relevant_sigs[i], interrupt);
	/* Set wakeup handler for the wakeup calls. */
	if (activeservice->rev_interval) {
	    if ( (pid = fork()) < 0 )
		error ("Fork failed: %s", strerror(errno));
	    else if (!pid) {
		set_program_title ("wakeup for %s", activeservice->name);
		wakeup_handler();
	    } else {
		msg ("Service %s: started wakeup handler at pid %d",
		     activeservice->name, pid);
		servicereport->rev_pid = pid;
	    }
	}
    }
    
    msg ("Service %s: awaiting activity on port %d (socket fd %d)",
	 activeservice->name, activeservice->port, server_sock);

    if (listen (server_sock, activeservice->backlog + 1) < 0)
	error ("Service %s: failed to listen to server_socket: %s",
	       activeservice->name, strerror(errno));

    /* Avoid zombies. */
    signal (SIGCHLD, SIG_IGN);

    /* We're a service now. Never return, never exit
     * (unless something is REALLY wrong).
     */
    while (1) {
	/* Block until we get a connection. */
	FD_ZERO (&set);
	FD_SET (server_sock, &set);
	if (select (FD_SETSIZE, &set, 0, 0, 0) < 0)
	    error ("Service %s: error while waiting for activity: %d (%s)",
		   activeservice->name, errno, strerror(errno));

	/* Accept the client-side connection. */
	size = sizeof(clientname);
	if ( (new = accept (server_sock, (struct sockaddr *) &clientname,
			    (socklen_t *) &size)) < 0 )
	    error ("Service %s: failure while accepting on server socket: %s",
		   activeservice->name, strerror(errno));

	/* Store client IP, it's used for lots of logging. */
	client_ip = inet_ntoa (clientname.sin_addr);
	msg ("Service %s: connection from %s, socket %d", 
	     activeservice->name, client_ip, new);

	/* Backend yet to be defined. */
	current_backend = -1;

	/* Incase of a http type service: handle separately. */
	if (activeservice->type == type_http) {
	    http_serve (new);
	    close (new);
	    continue;
	}
	
	/* This is an 'any' service type.
	 * Leave it alone if there are no back ends or if we exceed
	 * the max allowed clients. IN THAT CASE WE RETURN so that
	 * runservice() may close the listener socket, sleep, and
	 * create a new one.
	 */
	if (! backend_available()) {
	    warning ("Service %s: no back ends available",
		     activeservice->name);
	    close (new);
	    return;
	}

	/* Retry back ends until we succeed. */
	while (1) {
	    choose_backend();
	    if (current_backend < 0) {
		/* No back ends available now. NOTE: we return so that
		 * runservice() will retry in some time. */
		close (new);
		return;
	    }
	    
	    /* Show what's happening. */
	    msg ("Service %s: trying back end %s, port %d",
		 activeservice->name, 
		 activeservice->backend[current_backend].server,
		 activeservice->backend[current_backend].port);
	    
	    /* Connect to the backend. If this fails then we'll sleep
	     * and re-enter the while(1) loop instead of returning.
	     * In the loop we may need to decide to return after all. */
	    if ( (backend_sock = backend_connect()) < 0 ) {
		sysrun (activeservice->backend[current_backend].onfailure);
		warning ("Service %s: failed to connect to server %s:%d",
			 activeservice->name,
			 activeservice->backend[current_backend].server,
			 activeservice->backend[current_backend].port);
		continue;
	    }

	    /* Got a live one! */
	    if (fork_tcp_servicer (current_backend)) {
		/* We're the parent branch here. Close sockets so that
		 * we don't run out of file descriptors, and loop into
		 * the next select/accept run. */
		close (new);
		close (backend_sock);
		break;
	    }

	    /* Child branch: piggyback to and fro.
	     * This one never returns. */
	    log_activity_start ();
	    incr_client_count ();
	    flag_verbose = activeservice->backend[current_backend].verbosity;
	    copysockets (new, backend_sock);
	}
    }
}
